#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <dtrace.h>
#include <mach/mach.h>

/* API */

// based on this stack overflow question which explains the libdtrace API:
// https://stackoverflow.com/questions/28011495
// as well as the solaris libdtrace man page

static int chewrec(const dtrace_probedata_t* data, const dtrace_recdesc_t* rec, void* arg) {
  if (rec == NULL) {
    return DTRACE_CONSUME_NEXT;
  }

  dtrace_actkind_t act = rec->dtrd_action;

  if (act == DTRACEACT_EXIT) {
    return DTRACE_CONSUME_NEXT;
  }
  
  if (act == DTRACEACT_DIFEXPR) {
    // we'll define the result value as being trace((uint64_t))
    //printf("got DIFEXPR\n");
    uint32_t size = rec->dtrd_size;
    //printf("size is: %d\n", size);
    uint8_t* base = (uint8_t*)(data->dtpda_data);
    //printf("base: %p\n", base);

    if (size == 8 && base != NULL) {
      uint64_t val_u64 = *(uint64_t*)(base);
      // output it:
      //printf("outputing : 0x%016llx\n", val_u64);
      *(uint64_t*)arg = val_u64;
    }
    return DTRACE_CONSUME_NEXT;
  }

  return DTRACE_CONSUME_THIS;
}

static int chew(const dtrace_probedata_t* data, void* arg) {
  return DTRACE_CONSUME_THIS;
}

char* dtrace_helloworld_prog = "BEGIN { printf(\"Hello DTRACE\\n\"); trace((uint64_t)0x4141414142424242); exit(0);}";



static dtrace_hdl_t* dtrace_setup(char* script, int argc, char** argv) {
  int err = 0;

  /* open dtrace handle */

  dtrace_hdl_t* hdl = dtrace_open(DTRACE_VERSION, 0, &err);
  if (hdl == NULL) {
    printf("failed to open dtrace handle, err: 0x%x\n", err);
    exit(EXIT_FAILURE);
  }
  //printf("got dtrace handle: %p\n", hdl);

  /* set options */

  err = dtrace_setopt(hdl, "bufsize", "4m");
  if (err != 0) {
    printf("failed to set bufsize\n");
    exit(EXIT_FAILURE);
  }

  err = dtrace_setopt(hdl, "aggsize", "4m");
  if (err != 0) {
    printf("failed to set bufsize\n");
    exit(EXIT_FAILURE);
  }
  //printf("set options\n");

  /* compile program */

  dtrace_prog_t* prog = dtrace_program_strcompile(hdl,
                                                  script,
                                                  DTRACE_PROBESPEC_NAME,
                                                  0,
                                                  argc,
                                                  argv);
  if (prog == NULL) {
    printf("failed to compile program\n");
    printf("%s\n", dtrace_errmsg(hdl, dtrace_errno(hdl)));
    exit(EXIT_FAILURE);
  }
  //printf("compiled dtrace program\n");

  /* enable probes */

  dtrace_proginfo_t info;
  err = dtrace_program_exec(hdl, prog, &info);
  if (err != 0) {
    printf("failed to enable probes\n");
    exit(EXIT_FAILURE);
  }
  //printf("probes enabled\n");

  /* start instrumentation */

  err = dtrace_go(hdl);
  if (err != 0) {
    printf("failed to start instrumentation\n");
    exit(EXIT_FAILURE);
  }
  //printf("instrumentation started\n");
  
  return hdl;
}


static uint64_t dtrace_complete(dtrace_hdl_t* hdl) {
  int err = 0;

  uint64_t result_val = 0;

  while (1) {
    int done = 0;

    dtrace_sleep(hdl);

    // our result is a u64; pass the pointer in here:

    switch (dtrace_work(hdl, stdout, chew, chewrec, &result_val)) {
      case DTRACE_WORKSTATUS_DONE:
        done = 1;
        break;
      case DTRACE_WORKSTATUS_OKAY:
        break;
      default:
        printf("dtrace fatal error from dtrace_work\n");
        exit(EXIT_FAILURE);
    }

    if (done) {
      break;
    }
  }

  //printf("result val: 0x%016llx\n", result_val);

  err = dtrace_status(hdl);
  if (err == -1) {
    printf("dtrace_status failed\n");
    exit(EXIT_FAILURE);
  }

  err = dtrace_stop(hdl);
  if (err == -1) {
    printf("dtrace_stop failed\n");
    exit(EXIT_FAILURE);
  }

  err = dtrace_aggregate_print(hdl, stdout, NULL);
  if (err == -1) {
    printf("dtrace_aggregate_print failed\n");
    exit(EXIT_FAILURE);
  }

  dtrace_close(hdl);

  return result_val;
}

/*
  Run a dtrace script which returns a single u64 value by doing:
  trace((uint64_t)VAL);
  exit(0);
*/
static uint64_t run_dtrace_u64_trace_ret(char* script, int argc, char** argv) {
  
  dtrace_hdl_t* hdl = dtrace_setup(script, argc, argv);
  uint64_t result = dtrace_complete(hdl);
  return result;
}


// so this is still pretty inefficient as the argument is evaluated at compile time;
// if necessary could just mess with the compiled form?
uint64_t rk64(uint64_t kaddr) {
  printf("request to read: 0x%016llx\n", kaddr);
  char* kmemread_script = "BEGIN {trace((uint64_t)*(uint64_t*)$1); exit(0);}";
  char kaddr_str[32] = {0};
  sprintf(kaddr_str, "0x%llx", kaddr);
  char* argv[] = {"kmemread_script", kaddr_str, NULL};
  return run_dtrace_u64_trace_ret(kmemread_script, 2, argv);
}

uint32_t rk32(uint64_t kaddr) {
  return (uint32_t)rk64(kaddr);
}

uint64_t proc_pointer_for_pid(pid_t target_pid) {
  char* task_port_pointer_for_pid_script = 
"fbt:mach_kernel:proc_find:entry /arg0 == $1/ {self->match = 1;} \n"
"fbt:mach_kernel:proc_find:return /self->match/{trace((uint64_t)arg1); exit(0);}";
  
  char pid_str[32] = {0};
  sprintf(pid_str, "%d", target_pid);
  char* argv[] = {"task_pointer_script", pid_str, NULL};

  dtrace_hdl_t* hdl = dtrace_setup(task_port_pointer_for_pid_script, 2, argv);

  // now perform the action which that script should capture:
  mach_port_t target_port = MACH_PORT_NULL;
  kern_return_t kerr = task_for_pid(mach_task_self(), target_pid, &target_port);
  if (target_port) {
    mach_port_deallocate(mach_task_self(), target_port);
  }

  // capture the result:
  uint64_t result = dtrace_complete(hdl);
  return result;
}

#define OFFSET_PROC_TASK 0x10
#define OFFSET_TASK_SPACE 0x338
#define OFFSET_SPACE_SIZE 0xc
#define OFFSET_SPACE_TABLE 0x18

uint64_t lookup_port_in_proc(uint64_t proc, uint32_t port_name) {
  uint32_t port_index = port_name >> 8;

  uint64_t task = rk64(proc+OFFSET_PROC_TASK);
  printf("task: 0x%016llx\n", task);

  uint64_t space = rk64(task+OFFSET_TASK_SPACE);
  printf("space: 0x%016llx\n", space);


  uint32_t table_size = rk32(space+OFFSET_SPACE_SIZE);
  printf("table_size: 0x%08x\n", table_size);


  if (port_index >= table_size) {
    printf("port index out of bounds\n");
    return 0;
  }

  uint64_t table = rk64(space+OFFSET_SPACE_TABLE);

  uint64_t entry = table + (0x18 * port_index);
  uint64_t port = rk64(entry);

  printf("port: 0x%016llx\n", port);

  return port;
}

uint64_t port_kaddr(mach_port_t port) {
  return lookup_port_in_proc(proc_pointer_for_pid(getpid()), port);
}

#define OFFSET_PORT_MESSAGES 0x40

// dump the message queue:
void dump_port_message_queue(uint64_t port) {
  uint64_t messages = rk64(port + OFFSET_PORT_MESSAGES);
  printf("messages: 0x%016llx\n", messages);
}
#if 0
int main(int argc, char** argv) {
  //uint64_t kaddr = 0;
  //if (argc < 2) {
  //  printf("no kaddr\n");
  //  return 0;
  //}

  //kaddr = strtoull(argv[1], NULL, 16);

  //uint64_t val = rk64(kaddr);

  //printf("rk64 returned: 0x%016llx\n", val);

  uint64_t our_proc = proc_pointer_for_pid(getpid());
  printf("our_proc: 0x%016llx\n", our_proc);

  uint64_t task_self_port_pointer = lookup_port_in_proc(our_proc, mach_task_self());
  printf("task_self_port_pointer: 0x%016llx\n", task_self_port_pointer);

  uint64_t start_of_port = rk64(task_self_port_pointer);
  printf("%016llx\n", start_of_port);

  return 0;
}
#endif









